module Main exposing (..)

import Browser
import Csv.Decode as Decode exposing (Decoder)
import Html exposing (Html, b, button, div, text)
import Html.Attributes exposing (align)
import Html.Events exposing (onClick)
import Http
import List.Extra
import Model exposing (Model)
import ParallelCoordinates exposing (Msg(..))
import Result exposing (Result(..))
import Round
import Scatterplot exposing (Msg(..))
import SmartPhoneType exposing (Smartphone)
import Starplot exposing (Msg(..))
import String exposing (toInt)
import TreeView exposing (Msg(..))
import TypedSvg.Attributes exposing (class)
import Zoom



-- MAIN


main =
    Browser.element { init = init, update = update, subscriptions = subscriptions, view = view }


init : () -> ( Model, Cmd Msg )
init _ =
    ( Model []
        Nothing
        ""
        { attribute1 = "Price", attribute2 = "Rating", att1List = [], att2List = [], unit1 = "EUR", unit2 = "%" }
        { orderedList =
            [ .price
            , .rating
            , .num_cores
            , .processor_speed
            , .battery
            , .fast_charging
            , .memory
            , .ram
            , .screen_size
            , .refresh_rate
            ]
        , orderedLabels =
            [ "Price"
            , "Rating"
            , "NumCores"
            , "Processor Speed"
            , "Battery"
            , "Fast Charging"
            , "Memory"
            , "RAM"
            , "Screen Size"
            , "Refresh Rate"
            ]
        , orderedUnitLabels =
            [ "EUR"
            , "%"
            , ""
            , "GHz"
            , "mAh"
            , ""
            , "GB"
            , "GB"
            , "inches"
            , "Hz"
            ]
        , selectedFilter =
            { brand = "all"
            , g5 = "all"
            , nfc = "all"
            , ir_blaster = "all"
            }
        }
        { zoomOption =
            Zoom.init { width = TreeView.w, height = TreeView.h }
                |> Zoom.scaleExtent 1 100
                |> Zoom.translateExtent ( ( 0, 0 ), ( TreeView.w, TreeView.h ) )
        , dropCount = 0
        , inputDropCount = ""
        , selectedFilter =
            { brand = "all"
            , g5 = "true"
            , nfc = "false"
            , ir_blaster = "false"
            }
        }
        { modelID1 = "0"
        , modelID2 = "1019"
        , input1 = ""
        , input2 = ""
        , nextChange = 2
        }
        { scatterPlot = True
        , parallelCoordinatesPlot = False
        , treeView = False
        , starPlot = False
        }
    , fetchData
    )



-- UPDATE


type Msg
    = FetchData
    | DataReceived (Result Http.Error String)
    | ScatterplotMsg Scatterplot.Msg
    | ParallelCoordinatesMsg ParallelCoordinates.Msg
    | TreeViewMsg TreeView.Msg
    | StarplotMsg Starplot.Msg
    | ZooomMsg Zoom.OnZoom
    | SetViewScatterplot
    | SetViewParallelCoordesPlot
    | SetViewTree
    | SetViewStarPlot


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case msg of
        FetchData ->
            ( model, fetchData )

        DataReceived (Ok csvData) ->
            let
                parsedData =
                    parseCSV csvData
            in
            case parsedData of
                Ok data ->
                    ( { model | data = data, error = Nothing, csvdata = csvData }, Cmd.none )

                Err _ ->
                    ( { model | error = Just "Failed to parse CSV data" }, Cmd.none )

        DataReceived (Err _) ->
            ( { model | error = Just "Failed to load data" }, Cmd.none )

        ScatterplotMsg (ChangeAttribute1 value) ->
            let
                attList =
                    createAttList model value

                values =
                    Tuple.first attList

                unit =
                    Tuple.second attList
            in
            ( { model
                | scatterplotOptions =
                    { attribute1 = value
                    , attribute2 = model.scatterplotOptions.attribute2
                    , att1List = values
                    , att2List = model.scatterplotOptions.att2List
                    , unit1 = unit
                    , unit2 = model.scatterplotOptions.unit2
                    }
              }
            , Cmd.none
            )

        ScatterplotMsg (ChangeAttribute2 value) ->
            let
                attList =
                    createAttList model value

                values =
                    Tuple.first attList

                unit =
                    Tuple.second attList
            in
            ( { model
                | scatterplotOptions =
                    { attribute1 = model.scatterplotOptions.attribute1
                    , attribute2 = value
                    , att1List = model.scatterplotOptions.att1List
                    , att2List = values
                    , unit1 = model.scatterplotOptions.unit1
                    , unit2 = unit
                    }
              }
            , Cmd.none
            )

        ScatterplotMsg (SetStarPlot id) ->
            ( { model
                | starplotOption =
                    { modelID1 =
                        if model.starplotOption.nextChange == 1 then
                            id

                        else
                            model.starplotOption.modelID1
                    , modelID2 =
                        if model.starplotOption.nextChange == 2 then
                            id

                        else
                            model.starplotOption.modelID2
                    , input1 = model.starplotOption.input1
                    , input2 = model.starplotOption.input2
                    , nextChange =
                        if model.starplotOption.nextChange == 1 then
                            2

                        else
                            1
                    }
                , plotVisible = { scatterPlot = False, parallelCoordinatesPlot = False, treeView = False, starPlot = True }
              }
            , Cmd.none
            )

        ParallelCoordinatesMsg (ChangeOrder pos1 pos2) ->
            ( { model
                | parallelPlotOption =
                    { orderedList = List.Extra.swapAt pos1 pos2 model.parallelPlotOption.orderedList
                    , orderedLabels = List.Extra.swapAt pos1 pos2 model.parallelPlotOption.orderedLabels
                    , orderedUnitLabels = List.Extra.swapAt pos1 pos2 model.parallelPlotOption.orderedUnitLabels
                    , selectedFilter = model.parallelPlotOption.selectedFilter
                    }
              }
            , Cmd.none
            )

        ParallelCoordinatesMsg (SetScatterPlotFromParallelPlot att1 att2) ->
            let
                attList1 =
                    createAttList model att1

                attList2 =
                    createAttList model att2

                values1 =
                    Tuple.first attList1

                unit1 =
                    Tuple.second attList1

                values2 =
                    Tuple.first attList2

                unit2 =
                    Tuple.second attList2
            in
            ( { model
                | scatterplotOptions =
                    { attribute1 = att1
                    , attribute2 = att2
                    , att1List = values1
                    , att2List = values2
                    , unit1 = unit1
                    , unit2 = unit2
                    }
                , plotVisible = { scatterPlot = True, parallelCoordinatesPlot = False, treeView = False, starPlot = False }
              }
            , Cmd.none
            )

        ParallelCoordinatesMsg (SetBrandParallel value) ->
            ( { model
                | parallelPlotOption =
                    { orderedList = model.parallelPlotOption.orderedList
                    , orderedLabels = model.parallelPlotOption.orderedLabels
                    , orderedUnitLabels = model.parallelPlotOption.orderedUnitLabels
                    , selectedFilter =
                        { brand = value
                        , g5 = model.parallelPlotOption.selectedFilter.g5
                        , nfc = model.parallelPlotOption.selectedFilter.nfc
                        , ir_blaster = model.parallelPlotOption.selectedFilter.ir_blaster
                        }
                    }
              }
            , Cmd.none
            )

        ParallelCoordinatesMsg (Set5GParallel value) ->
            ( { model
                | parallelPlotOption =
                    { orderedList = model.parallelPlotOption.orderedList
                    , orderedLabels = model.parallelPlotOption.orderedLabels
                    , orderedUnitLabels = model.parallelPlotOption.orderedUnitLabels
                    , selectedFilter =
                        { brand = model.parallelPlotOption.selectedFilter.brand
                        , g5 = value
                        , nfc = model.parallelPlotOption.selectedFilter.nfc
                        , ir_blaster = model.parallelPlotOption.selectedFilter.ir_blaster
                        }
                    }
              }
            , Cmd.none
            )

        ParallelCoordinatesMsg (SetNFCParallel value) ->
            ( { model
                | parallelPlotOption =
                    { orderedList = model.parallelPlotOption.orderedList
                    , orderedLabels = model.parallelPlotOption.orderedLabels
                    , orderedUnitLabels = model.parallelPlotOption.orderedUnitLabels
                    , selectedFilter =
                        { brand = model.parallelPlotOption.selectedFilter.brand
                        , g5 = model.parallelPlotOption.selectedFilter.g5
                        , nfc = value
                        , ir_blaster = model.parallelPlotOption.selectedFilter.ir_blaster
                        }
                    }
              }
            , Cmd.none
            )

        ParallelCoordinatesMsg (SetIRBlasterParallel value) ->
            ( { model
                | parallelPlotOption =
                    { orderedList = model.parallelPlotOption.orderedList
                    , orderedLabels = model.parallelPlotOption.orderedLabels
                    , orderedUnitLabels = model.parallelPlotOption.orderedUnitLabels
                    , selectedFilter =
                        { brand = model.parallelPlotOption.selectedFilter.brand
                        , g5 = model.parallelPlotOption.selectedFilter.g5
                        , nfc = model.parallelPlotOption.selectedFilter.nfc
                        , ir_blaster = value
                        }
                    }
              }
            , Cmd.none
            )

        TreeViewMsg (ZoomMsg m) ->
            ( { model
                | treeOption =
                    { zoomOption = Zoom.update m model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter = model.treeOption.selectedFilter
                    }
              }
            , Cmd.none
            )

        TreeViewMsg (Click t) ->
            ( { model
                | treeOption =
                    { zoomOption = TreeView.performZoom t model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter = model.treeOption.selectedFilter
                    }
              }
            , Cmd.none
            )

        TreeViewMsg (SetStarPlotFromTree id) ->
            ( { model
                | starplotOption =
                    { modelID1 =
                        if model.starplotOption.nextChange == 1 then
                            id

                        else
                            model.starplotOption.modelID1
                    , modelID2 =
                        if model.starplotOption.nextChange == 2 then
                            id

                        else
                            model.starplotOption.modelID2
                    , input1 = model.starplotOption.input1
                    , input2 = model.starplotOption.input2
                    , nextChange =
                        if model.starplotOption.nextChange == 1 then
                            2

                        else
                            1
                    }
                , plotVisible = { scatterPlot = False, parallelCoordinatesPlot = False, treeView = False, starPlot = True }
              }
            , Cmd.none
            )

        TreeViewMsg (SetDropCountInput value) ->
            ( { model
                | treeOption =
                    { zoomOption = model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = value
                    , selectedFilter = model.treeOption.selectedFilter
                    }
              }
            , Cmd.none
            )

        TreeViewMsg SetDropCount ->
            ( { model
                | treeOption =
                    { zoomOption = model.treeOption.zoomOption
                    , dropCount = Maybe.withDefault 0 (String.toInt model.treeOption.inputDropCount)
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter = model.treeOption.selectedFilter
                    }
              }
            , Cmd.none
            )

        TreeViewMsg (SetBrandTree value) ->
            ( { model
                | treeOption =
                    { zoomOption = model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter =
                        { brand = value
                        , g5 = model.treeOption.selectedFilter.g5
                        , nfc = model.treeOption.selectedFilter.nfc
                        , ir_blaster = model.treeOption.selectedFilter.ir_blaster
                        }
                    }
              }
            , Cmd.none
            )

        TreeViewMsg (Set5GTree value) ->
            ( { model
                | treeOption =
                    { zoomOption = model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter =
                        { brand = model.treeOption.selectedFilter.brand
                        , g5 = value
                        , nfc = model.treeOption.selectedFilter.nfc
                        , ir_blaster = model.treeOption.selectedFilter.ir_blaster
                        }
                    }
              }
            , Cmd.none
            )

        TreeViewMsg (SetNFCTree value) ->
            ( { model
                | treeOption =
                    { zoomOption = model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter =
                        { brand = model.treeOption.selectedFilter.brand
                        , g5 = model.treeOption.selectedFilter.g5
                        , nfc = value
                        , ir_blaster = model.treeOption.selectedFilter.ir_blaster
                        }
                    }
              }
            , Cmd.none
            )

        TreeViewMsg (SetIRBlasterTree value) ->
            ( { model
                | treeOption =
                    { zoomOption = model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter =
                        { brand = model.treeOption.selectedFilter.brand
                        , g5 = model.treeOption.selectedFilter.g5
                        , nfc = model.treeOption.selectedFilter.nfc
                        , ir_blaster = value
                        }
                    }
              }
            , Cmd.none
            )

        ZooomMsg m ->
            ( { model
                | treeOption =
                    { zoomOption = Zoom.update m model.treeOption.zoomOption
                    , dropCount = model.treeOption.dropCount
                    , inputDropCount = model.treeOption.inputDropCount
                    , selectedFilter = model.treeOption.selectedFilter
                    }
              }
            , Cmd.none
            )

        StarplotMsg (InputChange1 id) ->
            ( { model
                | starplotOption =
                    { modelID1 = model.starplotOption.modelID1
                    , modelID2 = model.starplotOption.modelID2
                    , input1 = id
                    , input2 = model.starplotOption.input2
                    , nextChange = model.starplotOption.nextChange
                    }
              }
            , Cmd.none
            )

        StarplotMsg (InputChange2 id) ->
            ( { model
                | starplotOption =
                    { modelID1 = model.starplotOption.modelID1
                    , modelID2 = model.starplotOption.modelID2
                    , input1 = model.starplotOption.input1
                    , input2 = id
                    , nextChange = model.starplotOption.nextChange
                    }
              }
            , Cmd.none
            )

        StarplotMsg SubmitNumber1 ->
            ( { model
                | starplotOption =
                    { modelID1 =
                        case toInt model.starplotOption.input1 of
                            Just _ ->
                                model.starplotOption.input1

                            Nothing ->
                                model.starplotOption.modelID1
                    , modelID2 = model.starplotOption.modelID2
                    , input1 = model.starplotOption.input1
                    , input2 = model.starplotOption.input2
                    , nextChange =
                        if model.starplotOption.nextChange == 1 then
                            2

                        else
                            1
                    }
              }
            , Cmd.none
            )

        StarplotMsg SubmitNumber2 ->
            ( { model
                | starplotOption =
                    { modelID1 = model.starplotOption.modelID1
                    , modelID2 =
                        case toInt model.starplotOption.input2 of
                            Just _ ->
                                model.starplotOption.input2

                            Nothing ->
                                model.starplotOption.modelID2
                    , input1 = model.starplotOption.input1
                    , input2 = model.starplotOption.input2
                    , nextChange =
                        if model.starplotOption.nextChange == 1 then
                            2

                        else
                            1
                    }
              }
            , Cmd.none
            )

        SetViewScatterplot ->
            ( { model | plotVisible = { scatterPlot = True, parallelCoordinatesPlot = False, treeView = False, starPlot = False } }, Cmd.none )

        SetViewParallelCoordesPlot ->
            ( { model | plotVisible = { scatterPlot = False, parallelCoordinatesPlot = True, treeView = False, starPlot = False } }, Cmd.none )

        SetViewTree ->
            ( { model | plotVisible = { scatterPlot = False, parallelCoordinatesPlot = False, treeView = True, starPlot = False } }, Cmd.none )

        SetViewStarPlot ->
            ( { model | plotVisible = { scatterPlot = False, parallelCoordinatesPlot = False, treeView = False, starPlot = True } }, Cmd.none )


createAttList : Model -> String -> ( List (Maybe Float), String )
createAttList model att =
    if att == "Price" then
        ( List.map .price model.data, "EUR" )

    else if att == "Rating" then
        ( List.map .rating model.data, "%" )

    else if att == "NumCores" then
        ( List.map .num_cores model.data, "" )

    else if att == "Processor Speed" then
        ( List.map .processor_speed model.data, "GHz" )

    else if att == "Battery" then
        ( List.map .battery model.data, "mAh" )

    else if att == "Fast Charging" then
        ( List.map .fast_charging model.data, "" )

    else if att == "Memory" then
        ( List.map .memory model.data, "GB" )

    else if att == "RAM" then
        ( List.map .ram model.data, "GB" )

    else if att == "Screen Size" then
        ( List.map .screen_size model.data, "inches" )

    else if att == "Refresh Rate" then
        ( List.map .refresh_rate model.data, "Hz" )

    else
        ( [], "" )



-- VIEW


view : Model -> Html Msg
view model =
    div []
        [ div [ class [ "caption" ], align "center" ] [ b [] [ text "Smartphone Analysis Visualization" ] ]
        , div [ class [ "buttonMenu" ], align "center" ]
            [ button [ class [ "plotButton" ], onClick SetViewScatterplot ] [ b [] [ text "Scatterplot" ] ]
            , button [ class [ "plotButton" ], onClick SetViewParallelCoordesPlot ] [ b [] [ text "ParallelCoordinates" ] ]
            , button [ class [ "plotButton" ], onClick SetViewTree ] [ b [] [ text "TreeView" ] ]
            , button [ class [ "plotButton" ], onClick SetViewStarPlot ] [ b [] [ text "Starplot" ] ]
            ]
        , if model.plotVisible.scatterPlot then
            Scatterplot.viewScatterplot model |> Html.map ScatterplotMsg

          else
            div [] []
        , if model.plotVisible.parallelCoordinatesPlot then
            ParallelCoordinates.viewParallelCoordinates model |> Html.map ParallelCoordinatesMsg

          else
            div [] []
        , if model.plotVisible.treeView then
            TreeView.viewTree model |> Html.map TreeViewMsg

          else
            div [] []
        , if model.plotVisible.starPlot then
            Starplot.viewStarplot model |> Html.map StarplotMsg

          else
            div [] []
        ]


subscriptions : Model -> Sub Msg
subscriptions model =
    Zoom.subscriptions model.treeOption.zoomOption ZooomMsg



-- HTTP


fetchData : Cmd Msg
fetchData =
    Http.get
        { url = "./csv/smartphone_dataset.csv"
        , expect = Http.expectString DataReceived
        }



-- PARSING CSV


decoder : Decoder Smartphone
decoder =
    Decode.into Smartphone
        |> Decode.pipeline (Decode.field "id" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "brand" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "model" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "price" (Decode.blank Decode.float |> Decode.map toEuro))
        |> Decode.pipeline (Decode.field "rating" (Decode.blank Decode.float))
        |> Decode.pipeline (Decode.field "5G" (Decode.blank Decode.string |> Decode.map toBool))
        |> Decode.pipeline (Decode.field "NFC" (Decode.blank Decode.string |> Decode.map toBool))
        |> Decode.pipeline (Decode.field "IR_blaster" (Decode.blank Decode.string |> Decode.map toBool))
        |> Decode.pipeline (Decode.field "processor_name" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "processor_brand" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "num_cores" (Decode.blank Decode.string |> Decode.map toCores))
        |> Decode.pipeline (Decode.field "processor_speed" (Decode.blank Decode.string |> Decode.map toGHz))
        |> Decode.pipeline (Decode.field "Battery" (Decode.blank Decode.float))
        |> Decode.pipeline (Decode.field "fast_charging" (Decode.blank Decode.float))
        |> Decode.pipeline (Decode.field "Internal_mem" (Decode.blank Decode.string |> Decode.map toGB))
        |> Decode.pipeline (Decode.field "RAM" (Decode.blank Decode.float))
        |> Decode.pipeline (Decode.field "screen_size" (Decode.blank Decode.string |> Decode.map toInche))
        |> Decode.pipeline (Decode.field "resolution" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "refresh_rate" (Decode.blank Decode.float))
        |> Decode.pipeline (Decode.field "camera" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "card" (Decode.blank Decode.string))
        |> Decode.pipeline (Decode.field "os" (Decode.blank Decode.string))


parseCSV : String -> Result Decode.Error (List Smartphone)
parseCSV csv =
    Decode.decodeCsv Decode.FieldNamesFromFirstRow decoder csv



-- Umrechung Indische Rupie in Euro ungefähr


toEuro : Maybe Float -> Maybe Float
toEuro str =
    case str of
        Just price ->
            String.toFloat <| Round.round 2 <| price * 0.011

        _ ->
            Nothing


toBool : Maybe String -> Maybe Bool
toBool str =
    case str of
        Just "True" ->
            Just True

        Just "False" ->
            Just False

        _ ->
            Nothing


toCores : Maybe String -> Maybe Float
toCores str =
    case str of
        Just "dual core" ->
            Just 2

        Just "quad core" ->
            Just 4

        Just "hexa core" ->
            Just 6

        Just "octa core" ->
            Just 8

        _ ->
            Nothing


toGHz : Maybe String -> Maybe Float
toGHz str =
    str
        |> Maybe.withDefault ""
        |> String.replace "GHz Processor" ""
        |> String.trim
        |> String.toFloat


toGB : Maybe String -> Maybe Float
toGB str =
    str
        |> Maybe.withDefault ""
        |> String.replace "GB inbuilt" ""
        |> String.trim
        |> String.toFloat


toInche : Maybe String -> Maybe Float
toInche str =
    str
        |> Maybe.withDefault ""
        |> String.replace "inches" ""
        |> String.trim
        |> String.toFloat
