module Scatterplot exposing (..)

import Axis exposing (..)
import Html exposing (Html, option)
import Html.Attributes exposing (align, selected, value)
import Html.Events exposing (onClick, onInput)
import Model exposing (Model)
import Scale exposing (ContinuousScale)
import Statistics
import TypedSvg exposing (circle, g, svg, text_)
import TypedSvg.Attributes exposing (class, fontFamily, fontSize, textAnchor, transform, viewBox)
import TypedSvg.Attributes.InPx exposing (cx, cy, r, x, y)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (AnchorAlignment(..), Length(..), Transform(..))


w : Float
w =
    900


h : Float
h =
    430


padding : Float
padding =
    50


radius : Float
radius =
    2.0


tickCount : Int
tickCount =
    5


defaultExtent : ( number, number1 )
defaultExtent =
    ( 0, 200000 )


type Msg
    = ChangeAttribute1 String
    | ChangeAttribute2 String
    | SetStarPlot String


scatterplot : XyData -> Svg Msg
scatterplot model =
    let
        xValues : List Float
        xValues =
            List.map .x model.data

        yValues : List Float
        yValues =
            List.map .y model.data

        xScaleLocal : ContinuousScale Float
        xScaleLocal =
            xScale xValues

        yScaleLocal : ContinuousScale Float
        yScaleLocal =
            yScale yValues

        labelPositions : { x : Float, y : Float }
        labelPositions =
            { x = wideExtent xValues |> Tuple.second
            , y = wideExtent yValues |> Tuple.second
            }
    in
    svg [ viewBox 0 0 w h, TypedSvg.Attributes.width <| TypedSvg.Types.Percent 100, TypedSvg.Attributes.height <| TypedSvg.Types.Percent 100 ]
        [ TypedSvg.style [] [ TypedSvg.Core.text """
            .point circle { stroke: rgba(0, 0, 0,0.4); fill: rgba(255, 255, 255,0.3); }
            .point text { display: none; }
            .point:hover circle { stroke: rgba(0, 0, 0,1.0); fill: rgb(118, 214, 78); }
            .point:hover text { display: inline; }
          """ ]

        -- plot zeichnen x Achse und y Achse
        , g [ transform [ Translate (padding - 1) (h - padding - 1) ] ]
            [ xAxis xValues
            , text_
                [ x (Scale.convert xScaleLocal labelPositions.x - 20), y 30, textAnchor AnchorMiddle, fontSize <| Px 14 ]
                [ Html.text <|
                    model.xDescription
                        ++ (if String.length model.unitX > 0 then
                                " in "

                            else
                                ""
                           )
                        ++ model.unitX
                ]
            ]
        , g [ transform [ Translate (padding - 1) (padding - 1) ] ]
            [ yAxis yValues
            , text_
                [ x 20, y (Scale.convert yScaleLocal labelPositions.y - 10), textAnchor AnchorMiddle, fontSize <| Px 14 ]
                [ Html.text <|
                    model.yDescription
                        ++ (if String.length model.unitY > 0 then
                                " in "

                            else
                                ""
                           )
                        ++ model.unitY
                ]
            ]

        -- datenpunkte zeichnen
        , g [ transform [ Translate padding padding ] ]
            (List.map (point xScaleLocal yScaleLocal) model.data)
        ]


point : ContinuousScale Float -> ContinuousScale Float -> Point -> Svg Msg
point scaleX scaleY xyPoint =
    g [ class [ "point" ], fontSize <| Px 10.0, fontFamily [ "sans-serif" ], onClick (SetStarPlot xyPoint.id) ]
        [ circle
            [ cx (Scale.convert scaleX xyPoint.x)
            , cy (Scale.convert scaleY xyPoint.y)
            , r radius
            ]
            []
        , text_
            [ x (Scale.convert scaleX xyPoint.x)
            , y (Scale.convert scaleY xyPoint.y)
            , textAnchor AnchorMiddle
            ]
            [ Html.text ("ID:" ++ xyPoint.id ++ " | " ++ xyPoint.pointName) ]
        ]


xScale : List Float -> ContinuousScale Float
xScale values =
    Scale.linear ( 0, w - 2 * padding ) (wideExtent values)


yScale : List Float -> ContinuousScale Float
yScale values =
    Scale.linear ( h - 2 * padding, 0 ) (wideExtent values)


wideExtent : List Float -> ( Float, Float )
wideExtent values =
    let
        tup =
            Maybe.withDefault defaultExtent (Statistics.extent values)

        maxValue =
            Tuple.second tup + Tuple.second tup / (toFloat tickCount * 2)

        minValue =
            if (Tuple.first tup - Tuple.second tup / (toFloat tickCount * 2)) <= 0 then
                0

            else
                Tuple.first tup - Tuple.second tup / (toFloat tickCount * 2)
    in
    ( minValue, maxValue )


xAxis : List Float -> Svg msg
xAxis values =
    Axis.bottom [ Axis.tickCount tickCount ] (xScale values)


yAxis : List Float -> Svg msg
yAxis values =
    Axis.left [ Axis.tickCount tickCount ] (yScale values)


type alias Point =
    { pointName : String, id : String, x : Float, y : Float }


type alias XyData =
    { xDescription : String
    , yDescription : String
    , unitX : String
    , unitY : String
    , data : List Point
    }


filterSmartphonesXY : Model -> List (Maybe Float) -> List (Maybe Float) -> XyData
filterSmartphonesXY model att1List att2List =
    let
        modelList =
            List.map .model model.data

        idList =
            List.map .id model.data

        smartphoneDataList =
            List.map4 (\a1 a2 mn i -> { att1 = a1, att2 = a2, name = mn, id = i }) att1List att2List modelList idList

        filteredSmartphones =
            List.filterMap mapToPoint smartphoneDataList
    in
    XyData model.scatterplotOptions.attribute1
        model.scatterplotOptions.attribute2
        model.scatterplotOptions.unit1
        model.scatterplotOptions.unit2
        filteredSmartphones


mapToPoint : { att1 : Maybe Float, att2 : Maybe Float, name : Maybe String, id : Maybe String } -> Maybe Point
mapToPoint smartphoneData =
    Maybe.map4
        (\att1 att2 name index ->
            { pointName = name ++ " (" ++ String.fromFloat att1 ++ "," ++ String.fromFloat att2 ++ ")"
            , id = index
            , x = att1
            , y = att2
            }
        )
        smartphoneData.att1
        smartphoneData.att2
        smartphoneData.name
        smartphoneData.id


viewScatterplot : Model -> Html Msg
viewScatterplot model =
    let
        att1List =
            if model.scatterplotOptions.att1List == [] then
                List.map .price model.data

            else
                model.scatterplotOptions.att1List

        att2List =
            if model.scatterplotOptions.att2List == [] then
                List.map .rating model.data

            else
                model.scatterplotOptions.att2List

        filteredSmartphoneData =
            filterSmartphonesXY model att1List att2List
    in
    Html.div []
        [ Html.div [ class [ "description" ], align "center" ]
            [ Html.b [] [ Html.text "Hints: By clicking on a point, the plot switch to the star plot and show the point data. By hovering over a point, you get the model name, ID und the exactly x and y values." ]
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ Html.b [] [ Html.text "x-Axis: " ]
            , attributeSelector 1 model
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ Html.b [] [ Html.text " y-Axis: " ]
            , attributeSelector 2 model
            ]
        , scatterplot filteredSmartphoneData
        ]


attributeSelector : Int -> Model -> Html Msg
attributeSelector attNr model =
    let
        attributes =
            [ "Price"
            , "Rating"
            , "NumCores"
            , "Processor Speed"
            , "Battery"
            , "Fast Charging"
            , "Memory"
            , "RAM"
            , "Screen Size"
            , "Refresh Rate"
            ]

        att1 =
            model.scatterplotOptions.attribute1

        att2 =
            model.scatterplotOptions.attribute2
    in
    Html.select
        [ onInput
            (if attNr == 1 then
                ChangeAttribute1

             else
                ChangeAttribute2
            )
        ]
        (if attNr == 1 then
            List.map (createSelectorOptions att1) attributes

         else
            List.map (createSelectorOptions att2) attributes
        )


createSelectorOptions : String -> String -> Html Msg
createSelectorOptions select att =
    if select == att then
        option [ value att, selected True ] [ Html.text att ]

    else
        option [ value att ] [ Html.text att ]
