module Starplot exposing (..)

import Color
import Debug exposing (toString)
import Html exposing (Html, b, br, button, div, input, text)
import Html.Attributes exposing (align, placeholder, value)
import Html.Events exposing (onClick, onInput)
import List
import Model exposing (Model)
import Path
import Shape
import SmartPhoneType exposing (Smartphone)
import TypedSvg exposing (circle, g, svg, text_)
import TypedSvg.Attributes exposing (class, fill, stroke, viewBox)
import TypedSvg.Attributes.InPx exposing (cx, cy, r, strokeWidth, x, y)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Types exposing (AnchorAlignment(..), Length(..), Paint(..), Transform(..))


w : Float
w =
    990


h : Float
h =
    700


midX : Float
midX =
    450


midY : Float
midY =
    360


padding : Float
padding =
    50


starlength : Float
starlength =
    300


type Msg
    = InputChange1 String
    | SubmitNumber1
    | InputChange2 String
    | SubmitNumber2


starplot : Smartphone -> List Float -> List String -> List String -> List (Smartphone -> Maybe Float) -> Svg msg
starplot smartphone starValues labels unit func =
    let
        axisMultiplikator =
            [ ( 0, -0.9 )
            , ( 0.5, -0.75 )
            , ( 0.75, -0.25 )
            , ( 0.75, 0.25 )
            , ( 0.5, 0.75 )
            , ( 0, 0.9 )
            , ( -0.5, 0.75 )
            , ( -0.75, 0.25 )
            , ( -0.75, -0.25 )
            , ( -0.5, -0.75 )
            ]

        axisMultiplikatorHalf =
            List.map (\( x, y ) -> ( x / 2, y / 2 )) axisMultiplikator

        starAxisMax =
            List.map createAxis
                axisMultiplikator

        valueLabel =
            List.map
                (\f ->
                    case f smartphone of
                        Just value ->
                            value

                        Nothing ->
                            -1
                )
                func

        starLabel =
            List.map4 createLabels
                labels
                valueLabel
                unit
                [ ( 0, -1.1 )
                , ( 0.7, -0.95 )
                , ( 0.95, -0.45 )
                , ( 0.95, 0.45 )
                , ( 0.7, 0.95 )
                , ( 0, 1.1 )
                , ( -0.7, 0.95 )
                , ( -0.95, 0.45 )
                , ( -1.1, -0.45 ) -- individuel
                , ( -0.7, -0.95 )
                ]

        pathCircleMax =
            Shape.line Shape.linearCurve <|
                List.map (\( x, y ) -> Just ( midX + x * starlength, midY + y * starlength )) axisMultiplikator
                    ++ [ Just
                            ( midX + (Tuple.first <| Maybe.withDefault ( 0, 0.9 ) <| List.head axisMultiplikator) * starlength
                            , midY + (Tuple.second <| Maybe.withDefault ( 0, 0.9 ) <| List.head axisMultiplikator) * starlength
                            )
                       ]

        pathCircleHalf =
            Shape.line Shape.linearCurve <|
                List.map (\( x, y ) -> Just ( midX + x * starlength, midY + y * starlength )) axisMultiplikatorHalf
                    ++ [ Just
                            ( midX + (Tuple.first <| Maybe.withDefault ( 0, 0.9 ) <| List.head axisMultiplikatorHalf) * starlength
                            , midY + (Tuple.second <| Maybe.withDefault ( 0, 0.9 ) <| List.head axisMultiplikatorHalf) * starlength
                            )
                       ]

        pathCircleSmartphone =
            Shape.line Shape.linearCurve <|
                List.map2
                    (\axisM starV ->
                        Just
                            ( midX + Tuple.first axisM * starV
                            , midY + Tuple.second axisM * starV
                            )
                    )
                    axisMultiplikator
                    starValues
                    ++ [ Just
                            ( midX
                                + (Tuple.first <|
                                    Maybe.withDefault ( 0, 0.9 ) <|
                                        List.head axisMultiplikator
                                  )
                                * Maybe.withDefault 0 (List.head starValues)
                            , midY
                                + (Tuple.second <|
                                    Maybe.withDefault ( 0, 0.9 ) <|
                                        List.head axisMultiplikator
                                  )
                                * Maybe.withDefault 0 (List.head starValues)
                            )
                       ]
    in
    svg
        [ viewBox 0 0 w h
        , TypedSvg.Attributes.width <| TypedSvg.Types.Percent 50
        , TypedSvg.Attributes.height <| TypedSvg.Types.Percent 100
        ]
        ([ TypedSvg.style [] [ TypedSvg.Core.text """
            .valueLabel { display: none }
            .starLabel {font-size:30px;}
            .valueLabel {font-size:30px;}
            .label:hover .valueLabel { display: inline; }
            .label:hover .starLabel { display: none; }
            """ ]
         , circle
            [ cx midX
            , cy midY
            , r 1
            , stroke (Paint Color.black)
            , strokeWidth 1
            ]
            []
         , Path.element pathCircleMax
            [ stroke <| Paint <| Color.darkGray
            , strokeWidth 0.8
            , fill PaintNone
            ]
         , Path.element pathCircleHalf
            [ stroke <| Paint <| Color.darkGray
            , strokeWidth 0.5
            , fill PaintNone
            ]
         , Path.element pathCircleSmartphone
            [ stroke <| Paint <| Color.red
            , strokeWidth 2
            , fill PaintNone
            ]
         ]
            ++ starLabel
            ++ List.map
                (\p ->
                    Path.element p
                        [ stroke <| Paint <| Color.black
                        , strokeWidth 0.5
                        , fill PaintNone
                        ]
                )
                starAxisMax
        )


createAxis : ( Float, Float ) -> Path.Path
createAxis axisPoint =
    [ Just ( midX, midY ), Just ( midX + Tuple.first axisPoint * starlength, midY + Tuple.second axisPoint * starlength ) ] |> Shape.line Shape.linearCurve


createLabels : String -> Float -> String -> ( Float, Float ) -> Svg msg
createLabels label vlabel unit labelPoint =
    g [ class [ "label" ] ]
        [ g [ class [ "starLabel" ] ]
            [ text_ [ x (midX + Tuple.first labelPoint * starlength - 20), y (midY + Tuple.second labelPoint * starlength) ] [ Html.text label ]
            ]
        , g [ class [ "valueLabel" ] ]
            [ text_ [ x (midX + Tuple.first labelPoint * starlength - 20), y (midY + Tuple.second labelPoint * starlength) ]
                [ Html.text <|
                    if vlabel < 0 then
                        "NaN"

                    else
                        toString vlabel ++ " " ++ unit
                ]
            ]
        ]


findSmartphoneById : Model -> String -> Maybe Smartphone
findSmartphoneById model id =
    List.head <|
        List.filter (checkID id) model.data


findMaxValue : Model -> (Smartphone -> Maybe Float) -> Maybe Float
findMaxValue model func =
    List.maximum <| List.filterMap func model.data


checkID : String -> Smartphone -> Bool
checkID modelID smartphone =
    Maybe.withDefault "NaN" smartphone.id == modelID


calculateStarValue : Maybe Smartphone -> Maybe Float -> (Smartphone -> Maybe Float) -> Float
calculateStarValue smartphone max func =
    case smartphone of
        Just sp ->
            case func sp of
                Just sV ->
                    case max of
                        Just m ->
                            sV / m * starlength

                        Nothing ->
                            -1

                Nothing ->
                    -1

        Nothing ->
            -1


viewStarplot : Model -> Html Msg
viewStarplot model =
    let
        smartphoneFunctions =
            [ .price
            , .rating
            , .num_cores
            , .processor_speed
            , .battery
            , .fast_charging
            , .memory
            , .ram
            , .screen_size
            , .refresh_rate
            ]

        labels =
            [ "Price"
            , "Rating"
            , "NumCores"
            , "Processor Speed"
            , "Battery"
            , "Fast Charging"
            , "Memory"
            , "RAM"
            , "Screen Size"
            , "Refresh Rate"
            ]

        unit =
            [ "EUR"
            , "%"
            , ""
            , "GHz"
            , "mAh"
            , ""
            , "GB"
            , "GB"
            , "inches"
            , "Hz"
            ]

        starplotData1 =
            findSmartphoneById model model.starplotOption.modelID1

        starplotData2 =
            findSmartphoneById model model.starplotOption.modelID2

        maxValues =
            List.map (findMaxValue model)
                smartphoneFunctions

        starValues1 =
            List.map2 (calculateStarValue starplotData1)
                maxValues
                smartphoneFunctions

        starValues2 =
            List.map2 (calculateStarValue starplotData2)
                maxValues
                smartphoneFunctions

        starplotViewHead spd1 spd2 =
            div []
                [ Html.div [ class [ "description" ], align "center" ] [ Html.b [] [ Html.text "Hints: You can compare two smartphones at the same time. Each of the two starplots has an input field where an ID from a smartphone can be entered to display it in the starplot. By hovering over a plot label, you can see the exactly value for the label." ] ]
                , Html.span [ class [ "plotStarInfo" ] ]
                    [ case spd1 of
                        Just spd ->
                            Html.span []
                                [ b [] [ text <| ("Starplot1 | ID:" ++ Maybe.withDefault "NaN" spd.id) ++ (" | " ++ Maybe.withDefault "NaN" spd.model) ]
                                , br [] []
                                , b [] [ text "Select ID1: " ]
                                , input [ placeholder "Enter a number", onInput InputChange1 ] []
                                , button [ onClick SubmitNumber1 ] [ text "Submit for SP1" ]
                                ]

                        Nothing ->
                            Html.span []
                                [ b [] [ text <| ("Starplot1 | ID:" ++ "NaN") ++ (" | " ++ "NaN") ]
                                , br [] []
                                , b [] [ text "Select ID1: " ]
                                , input [ placeholder "Enter a number", onInput InputChange1 ] []
                                , button [ onClick SubmitNumber1 ] [ text "Submit for SP1" ]
                                ]
                    , case spd2 of
                        Just spd ->
                            Html.span []
                                [ b [] [ text <| ("Starplot2 | ID:" ++ Maybe.withDefault "NaN" spd.id) ++ (" | " ++ Maybe.withDefault "NaN" spd.model) ]
                                , br [] []
                                , b [] [ text " Select ID2: " ]
                                , input [ placeholder "Enter a number", onInput InputChange2 ] []
                                , button [ onClick SubmitNumber2 ] [ text "Submit for SP2" ]
                                ]

                        Nothing ->
                            Html.span []
                                [ b [] [ text <| ("Starplot2 | ID:" ++ "NaN") ++ (" | " ++ "NaN") ]
                                , br [] []
                                , b [] [ text " Select ID2: " ]
                                , input [ placeholder "Enter a number", onInput InputChange2 ] []
                                , button [ onClick SubmitNumber2 ] [ text "Submit for SP2" ]
                                ]
                    ]
                ]
    in
    div []
        [ starplotViewHead starplotData1 starplotData2
        , case starplotData1 of
            Just spd1 ->
                starplot spd1 starValues1 labels unit smartphoneFunctions

            Nothing ->
                div []
                    [ text <| "Error: Starplot1 kann nicht dargestellt werden, da zu dieser ID (" ++ model.starplotOption.modelID1 ++ ") keine Data vorliegen Probiere Sie eine andere ID aus. Die ID muss eine Nummer sein."
                    ]
        , case starplotData2 of
            Just spd2 ->
                starplot spd2 starValues2 labels unit smartphoneFunctions

            Nothing ->
                div []
                    [ text <| "Error: Starplot2 kann nicht dargestellt werden, da zu dieser ID (" ++ model.starplotOption.modelID2 ++ ") keine Data vorliegen Probiere Sie eine andere ID aus. Die ID muss eine Nummer sein."
                    ]
        ]
