module TreeView exposing (..)

import Color
import Curve
import Debug exposing (toString)
import Hierarchy
import Html exposing (Html, b, button, div, input, option, select, text)
import Html.Attributes exposing (align, selected, value)
import Html.Events exposing (onClick, onInput)
import List.Extra
import Model exposing (Model)
import Path
import Set
import Shape
import SmartPhoneType exposing (Smartphone)
import Tree exposing (Tree)
import TypedSvg exposing (g, rect, svg)
import TypedSvg.Attributes exposing (class, dy, fill, href, id, pointerEvents, stroke, style, textAnchor, transform, viewBox)
import TypedSvg.Attributes.InPx exposing (fontSize, height, strokeWidth, width, x, y)
import TypedSvg.Core exposing (Svg)
import TypedSvg.Events
import TypedSvg.Types exposing (AnchorAlignment(..), ClipPath(..), Paint(..), Transform(..), em)
import Zoom exposing (Zoom)


w : Float
w =
    990


h : Float
h =
    450


padding : Float
padding =
    30


layedOut : Model -> ( Tree Datum, ( Int, Int ) )
layedOut model =
    let
        treeModelInfo =
            tree model

        treeModel =
            Tuple.first treeModelInfo

        treeNodeCounts =
            Tuple.second treeModelInfo
    in
    ( Hierarchy.tidy
        [ Hierarchy.nodeSize
            (\{ kind, size } ->
                case kind of
                    Leaf ->
                        ( sqrt size / 1.1, sqrt size * 2.5 )

                    Node ->
                        ( sqrt size * 1.7, sqrt size / 1.1 )
            )
        , Hierarchy.parentChildMargin 900
        , Hierarchy.peerMargin 60
        , Hierarchy.size (w - padding * 2) (h - padding * 2)
        ]
        treeModel
        |> Tree.sumUp
            (\n ->
                { x = n.x
                , y = n.y
                , width = n.width
                , height = n.height
                , node = n.node
                , bbox = { x = n.x, y = n.y, width = n.width, height = n.height }
                }
            )
            (\n c ->
                { x = n.x
                , y = n.y
                , width = n.width
                , height = n.height
                , node = n.node
                , bbox = List.foldl (\item t -> maxBBoxes item.bbox t) { x = n.x, y = n.y, width = n.width, height = n.height } c
                }
            )
    , treeNodeCounts
    )


maxBBoxes :
    { x : Float, y : Float, width : Float, height : Float }
    -> { x : Float, y : Float, width : Float, height : Float }
    -> { x : Float, y : Float, width : Float, height : Float }
maxBBoxes a b =
    let
        x =
            min a.x b.x

        y =
            min a.y b.y
    in
    { x = x
    , y = y
    , width = max (a.x + a.width) (b.x + b.width) - x
    , height = max (a.y + a.height) (b.y + b.height) - y
    }


type alias Datum =
    { x : Float
    , y : Float
    , width : Float
    , height : Float
    , node : { size : Float, id : String, name : String, kind : Kind }
    , bbox : { x : Float, y : Float, width : Float, height : Float }
    }


type Msg
    = ZoomMsg Zoom.OnZoom
    | Click Datum
    | SetStarPlotFromTree String
    | SetDropCountInput String
    | SetDropCount
    | SetBrandTree String
    | Set5GTree String
    | SetNFCTree String
    | SetIRBlasterTree String


performZoom : Datum -> Zoom -> Zoom
performZoom t =
    let
        p =
            ( t.x + t.width / 2, t.y + t.height / 2 )

        xScale =
            (w - 2 * padding) / t.bbox.width

        yScale =
            (h - 2 * padding) / t.bbox.height

        scale =
            min xScale yScale

        tx =
            (padding + t.bbox.x) * -scale

        ty =
            (padding + t.bbox.y) * -scale

        translate =
            if xScale > yScale then
                { x = w / 2 - scale * t.bbox.width / 2 + tx, y = padding + ty }

            else
                { x = padding + tx, y = h / 2 - scale * t.bbox.height / 2 + ty }
    in
    Zoom.setTransform (Zoom.animatedAround p) { scale = scale, translate = translate }


viewTree : Model -> Html Msg
viewTree model =
    let
        brandList =
            Set.toList <| Set.fromList <| List.filterMap .brand model.data

        brandFilterValues =
            "all" :: brandList

        otherFilterValues =
            [ "all", "true", "false" ]

        treeViewInfo =
            createViewTree model

        treeView =
            Tuple.first treeViewInfo

        nodeCounts =
            Tuple.second treeViewInfo
    in
    div []
        [ Html.div [ class [ "description" ], align "center" ] [ Html.b [] [ Html.text "Hints: By scrolling you can zoom out or in the visualization and you can move the tree. By clicking on a node (not a leaf), the tree focuses and zooms on this node. By clicking on a leaf node, you switch to the starplot of the node. You can select different filters to bound the result." ] ]
        , Html.div [ class [ "plotTreeInfo" ] ]
            [ Html.span [ class [ "plotAttributeSelect" ] ]
                [ b [] [ text <| "Smartphones in Tree: " ++ (toString <| Tuple.second nodeCounts) ]
                ]
            , Html.span [ class [ "plotAttributeSelect" ] ]
                [ b [] [ text <| "Total Smartphones: " ++ (toString <| Tuple.first nodeCounts) ]
                ]
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text " Brand: " ]
            , select [ onInput SetBrandTree ] (List.map (createSelectorOptions model.treeOption.selectedFilter.brand) brandFilterValues)
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text " 5G:  " ]
            , select [ onInput Set5GTree ]
                (List.map (createSelectorOptions model.treeOption.selectedFilter.g5) otherFilterValues)
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text " NFC: " ]
            , select [ onInput SetNFCTree ]
                (List.map (createSelectorOptions model.treeOption.selectedFilter.nfc) otherFilterValues)
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text " IR-Blaster: " ]
            , select [ onInput SetIRBlasterTree ]
                (List.map (createSelectorOptions model.treeOption.selectedFilter.ir_blaster) otherFilterValues)
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text "Number of nodes to remove: " ]
            , input [ onInput SetDropCountInput ] []
            , button [ onClick SetDropCount ] [ text "Remove nodes" ]
            ]
        , Html.span [ class [ "plotAttributeSelect" ] ]
            [ b [] [ text ("Manual removed nodes: " ++ toString model.treeOption.dropCount) ]
            ]
        , treeView
        ]


createSelectorOptions : String -> String -> Html Msg
createSelectorOptions select valueOption =
    if select == valueOption then
        option [ value valueOption, selected True ] [ Html.text valueOption ]

    else
        option [ value valueOption ] [ Html.text valueOption ]


createViewTree : Model -> ( Svg Msg, ( Int, Int ) )
createViewTree model =
    let
        layoutTreeInfo =
            layedOut model

        layoutTree =
            Tuple.first layoutTreeInfo

        nodeCounts =
            Tuple.second layoutTreeInfo
    in
    ( svg
        [ viewBox 0 0 w h
        , TypedSvg.Attributes.width <| TypedSvg.Types.Percent 100
        , TypedSvg.Attributes.height <| TypedSvg.Types.Percent 100
        ]
        [ rect
            (width w
                :: height h
                :: fill (Paint (Color.rgba 0 0 0 0))
                :: stroke (Paint Color.black)
                :: strokeWidth 2
                :: Zoom.events model.treeOption.zoomOption ZoomMsg
            )
            []
        , g [ Zoom.transform model.treeOption.zoomOption ]
            [ g [ transform [ Translate padding padding ] ]
                [ layoutTree
                    |> Tree.links
                    |> List.map
                        (\( from, to ) ->
                            Shape.bumpYCurve
                                [ ( from.x + from.width / 2, from.y + from.height )
                                , ( to.x + to.width / 2, to.y + to.height * 0.1 )
                                ]
                        )
                    |> (\p ->
                            Path.element p
                                [ fill PaintNone
                                , stroke (Paint (Color.rgb 0.3 0.3 0.3))
                                , style "vector-effect: non-scaling-stroke"
                                , pointerEvents "none"
                                ]
                       )
                , layoutTree
                    |> Tree.toList
                    |> List.map label
                    |> g []
                ]
            ]
        ]
    , nodeCounts
    )


label : Datum -> Svg Msg
label item =
    case item.node.kind of
        Leaf ->
            g [ transform [ Translate item.x item.y ], onClick (SetStarPlotFromTree item.node.id) ]
                [ Path.element
                    [ Curve.linearClosed
                        [ ( 0, item.height )
                        , ( 0, 0 )
                        , ( item.width, 0 )
                        , ( item.width, item.height )
                        ]
                    ]
                    [ fill (Paint (Color.rgb 1 1 0.8))
                    , stroke (Paint Color.black)
                    , style "vector-effect: non-scaling-stroke"
                    , id ("path-" ++ item.node.name)
                    ]
                , TypedSvg.title [] [ TypedSvg.Core.text (item.node.name ++ " - " ++ String.fromFloat item.node.size ++ " loc") ]
                , TypedSvg.clipPath [ id ("clip-" ++ item.node.name) ]
                    [ TypedSvg.use [ href ("#path-" ++ item.node.name) ] []
                    ]
                , item.node.name
                    |> String.split "."
                    |> List.reverse
                    |> List.head
                    |> Maybe.withDefault item.node.name
                    |> String.foldl
                        (\c ( word, soFar ) ->
                            if Char.toCode c == 32 then
                                ( [ c ]
                                , case word of
                                    [] ->
                                        soFar

                                    _ ->
                                        String.fromList (List.reverse word) :: soFar
                                )

                            else
                                ( c :: word, soFar )
                        )
                        ( [], [] )
                    |> (\( word, soFar ) -> String.fromList (List.reverse word) :: soFar)
                    |> List.reverse
                    |> List.indexedMap
                        (\i st ->
                            TypedSvg.tspan
                                [ x (item.width * 0.05)
                                , TypedSvg.Attributes.y (em (2 + toFloat i * 0.9))
                                ]
                                [ TypedSvg.Core.text st ]
                        )
                    |> TypedSvg.text_
                        [ fontSize (sqrt item.node.size / 130)
                        , pointerEvents "none"
                        , TypedSvg.Attributes.clipPath (ClipPathFunc ("url(#clip-" ++ item.node.name ++ ")"))
                        ]
                ]

        Node ->
            g []
                [ Path.element
                    [ Curve.linearClosed
                        [ ( item.x, item.y + item.height )
                        , ( item.x, item.y + item.height * 0.1 )
                        , ( item.x + item.width * 0.45, item.y + item.height * 0.1 )
                        , ( item.x + item.width, item.y + item.height * 0.1 )
                        , ( item.x + item.width, item.y + item.height )
                        ]
                    ]
                    [ fill (Paint (Color.rgb 0.8 0.9 1)) -- blue
                    , stroke (Paint Color.black)
                    , style "vector-effect: non-scaling-stroke"
                    , TypedSvg.Events.onClick (Click item)
                    ]
                , TypedSvg.title [] [ TypedSvg.Core.text (item.node.name ++ " - " ++ String.fromFloat item.node.size ++ " loc") ]
                , TypedSvg.text_
                    [ textAnchor AnchorMiddle
                    , fontSize (sqrt item.node.size / 110)
                    , pointerEvents "none"
                    , x (item.x + item.width / 2)
                    , y (item.y + item.height / 2)
                    , dy (em 0.3)
                    ]
                    [ TypedSvg.Core.text (String.split "." item.node.name |> List.Extra.last |> Maybe.withDefault "") ]
                ]


type Kind
    = Leaf
    | Node


tree : Model -> ( Tree { name : String, id : String, kind : Kind, size : Float }, ( Int, Int ) )
tree model =
    let
        filteredSmartphoneData =
            filterSPData model model.data

        notfilteredSmartphoneNodeCount =
            List.length model.data

        filteredSmartphoneNodeCount =
            List.length filteredSmartphoneData

        limitedfilteredSmartphones =
            List.drop model.treeOption.dropCount filteredSmartphoneData
    in
    ( Tree.stratifyWithPath
        { path = \item -> String.split "." item.name
        , createMissingNode = \path -> { name = String.join "." path, id = "-1", size = 0 }
        }
        (getTreeData limitedfilteredSmartphones)
        |> Result.withDefault (Tree.singleton { name = "Smartphone", id = "-1", size = 0 })
        |> Tree.sumUp (\n -> { name = n.name, id = n.id, size = n.size, kind = Leaf })
            (\node children ->
                { name = node.name, id = node.id, size = List.sum (List.map .size children), kind = Node }
            )
        |> Tree.sortBy (Tree.label >> .name)
    , ( notfilteredSmartphoneNodeCount, filteredSmartphoneNodeCount )
    )


getTreeData : List Smartphone -> List { name : String, id : String, size : Float }
getTreeData smartphones =
    let
        listlength =
            List.length smartphones
    in
    List.map (getTreeString listlength) smartphones


getTreeString : Int -> Smartphone -> { name : String, id : String, size : Float }
getTreeString nodeCount smartphone =
    { name = "Smartphone." ++ Maybe.withDefault "" smartphone.brand ++ "." ++ Maybe.withDefault "" smartphone.model
    , id = Maybe.withDefault "-1" smartphone.id
    , size =
        if nodeCount < 20 then
            1000000

        else if nodeCount >= 20 && nodeCount < 40 then
            500000

        else if nodeCount >= 40 && nodeCount < 70 then
            200000

        else if nodeCount >= 70 && nodeCount < 100 then
            50000

        else if nodeCount >= 100 && nodeCount < 200 then
            20000

        else if nodeCount >= 200 && nodeCount < 300 then
            5000

        else if nodeCount >= 300 && nodeCount < 500 then
            1000

        else
            500
    }


filterSPData : Model -> List Smartphone -> List Smartphone
filterSPData model smartphones =
    let
        smartphoneList =
            List.filter (hasIRBlaster model) <|
                List.filter (hasNFC model) <|
                    List.filter (has5G model) <|
                        List.filter (fromBrand model) <|
                            smartphones
    in
    smartphoneList


fromBrand : Model -> Smartphone -> Bool
fromBrand model smartphone =
    if model.treeOption.selectedFilter.brand == Maybe.withDefault "" smartphone.brand || model.treeOption.selectedFilter.brand == "all" then
        True

    else
        False


has5G : Model -> Smartphone -> Bool
has5G model smartphone =
    if
        (model.treeOption.selectedFilter.g5 == "true" && Maybe.withDefault False smartphone.g5)
            || model.treeOption.selectedFilter.g5
            == "all"
    then
        True

    else if model.treeOption.selectedFilter.g5 == "false" && not (Maybe.withDefault True smartphone.g5) then
        True

    else
        False


hasNFC : Model -> Smartphone -> Bool
hasNFC model smartphone =
    if
        (model.treeOption.selectedFilter.nfc == "true" && Maybe.withDefault False smartphone.nfc)
            || model.treeOption.selectedFilter.nfc
            == "all"
    then
        True

    else if model.treeOption.selectedFilter.nfc == "false" && not (Maybe.withDefault True smartphone.nfc) then
        True

    else
        False


hasIRBlaster : Model -> Smartphone -> Bool
hasIRBlaster model smartphone =
    if
        (model.treeOption.selectedFilter.ir_blaster == "true" && Maybe.withDefault False smartphone.ir_blaster)
            || model.treeOption.selectedFilter.ir_blaster
            == "all"
    then
        True

    else if model.treeOption.selectedFilter.ir_blaster == "false" && not (Maybe.withDefault True smartphone.ir_blaster) then
        True

    else
        False
